from numpy import ndarray
from typing_extensions import Self
from sklearn.preprocessing import LabelEncoder
import numpy as np
import tensorflow as tf
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Dropout
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.metrics import F1Score
from tensorflow.keras.utils import to_categorical
from helical.models.base_models import BaseTaskModel
from pathlib import Path

class NeuralNetwork(BaseTaskModel):
    def __init__(self, loss: str = "categorical_crossentropy", learning_rate: float = 0.001, epochs=10, batch_size=32) -> None:
        self.loss = loss
        self.learning_rate = learning_rate
        self.epochs = epochs
        self.batch_size = batch_size
        self.encoder = LabelEncoder()

    def compile(self, num_classes: int, input_shape: int) -> None:
        """Compile a neural network. The model is a simple feedforward neural network with 2 hidden layers. 
        TODO - Add more flexibility to the model architecture.
        
        Parameters
        ----------
        num_classes : int
            The number of classes to predict.
        input_shape : int
            The input shape of the neural network.
        """

        self.num_classes = num_classes
        self.input_shape = (input_shape,)

        self.model = Sequential()
        self.model.add(Dense(256, activation='relu', input_shape=self.input_shape))
        self.model.add(Dropout(0.4))
        self.model.add(Dense(64, activation='relu'))
        self.model.add(Dropout(0.4))
        self.model.add(Dense(self.num_classes, activation='softmax'))

        self.optimizer = Adam(learning_rate=self.learning_rate)
        self.f1_metric = F1Score(average='macro')
        self.model.compile(loss=self.loss, optimizer=self.optimizer, metrics=[self.f1_metric])

    def train(self, X_train: ndarray, y_train: ndarray, validation_data: tuple[ndarray, ndarray]) -> Self:
        """Train the neural network on the training and validation data.

        Parameters
        ----------
        X_train : ndarray
            The training data features.
        y_train : ndarray
            The training data labels.
        validation_data : tuple(ndarray, ndarray)
            The validation data features and labels.

        Returns
        -------
        The neural network instance.
        """
        x_val, y_val = validation_data
        self.encoder.fit_transform(np.concatenate((y_train, y_val), axis = 0))

        y_train_encoded = self.encoder.transform(y_train)
        y_train_encoded = to_categorical(y_train_encoded, num_classes = self.num_classes)

        y_val_encoded = self.encoder.transform(y_val)
        y_val_encoded = to_categorical(y_val_encoded, num_classes = self.num_classes)

        self.model.fit(X_train, y_train_encoded, epochs = self.epochs, batch_size = self.batch_size, validation_data = (x_val, y_val_encoded))
        return self
    
    def predict(self, x: ndarray) -> ndarray:
        """Use the neural network to make predictions.

        Parameters
        ----------
        x : ndarray
            The data to make predictions upon.

        Returns
        -------
        The prediction of the neural network.
        """
        predictions = self.model.predict(x)
        y_pred = np.argmax(predictions, axis=1)
        return self.encoder.inverse_transform(y_pred)
    
    def save(self, path: str) -> None:
        """Save the neural network model and its encoder to a directory.
        Any missing parents of this path are created as needed.

        Parameters
        ----------
        path : str
            The path to the directory to save the model and the encoder.
        """
        Path(path).mkdir(parents=True, exist_ok=True)
        np.save(f"{path}/encoder", self.encoder.classes_)
        self.model.save(f"{path}/neural_network.h5")

    def load(self, path: str, classes: ndarray) -> Self:
        """Load the neural network model from a file.

        Parameters
        ----------
        path : str
            The path to load the model from.
        classes : ndarray
            The classes used for encoding the labels.

        Returns
        -------
        The neural network instance.
        """
        # set to None, showing this is a loaded model and not trained
        self.loss = None
        self.learning_rate = None
        self.epochs = None
        self.batch_size = None

        self.encoder.classes_ = np.load(classes)
        self.model = tf.keras.models.load_model(path)
        return self




from sklearn.model_selection import train_test_split
from helical.models.base_models import BaseTaskModel, BaseModelProtocol
import logging
import numpy as np
from numpy import ndarray
from anndata import AnnData
from typing import Protocol, runtime_checkable, Optional, Union, Optional
from typing_extensions import Self

@runtime_checkable
class ClassificationModelProtocol(Protocol):
    def predict(self, x: Union[AnnData, ndarray]) -> ndarray:
        ...

LOGGER = logging.getLogger(__name__)

class Classifier():
    def __init__(self) -> None:

        self.trained_task_model = None
        self.base_model = None
        self.name = None
        self.gene_names = None

    def get_predictions(self, x: Union[AnnData, ndarray], gene_names: Optional[str] = None) -> ndarray:
        """
        Make predictions on the data. 
        
        - If no base model is provided, it is assumed that the trained_task_model is a custom 
            standalone model that can handle the data directly.
        - If a base model is provided, the data is processed by the base model and the embeddings are used as input to the trained_task_model.

        Parameters
        ----------
        x : AnnData
            The predictions for each model.
        gene_names : str, optional
            The name of the column in the var attribute of the AnnData object that contains the gene names.
            If none is provided, the member variable self.gene_names is used.

        Returns
        -------
        A numpy array with the predictions.
        """
        if self.base_model is not None: 
            if gene_names:
                self.gene_names = gene_names
            dataset = self.base_model.process_data(x, self.gene_names)
            embeddings = self.base_model.get_embeddings(dataset)
            x = embeddings
        
        return self.trained_task_model.predict(x) 

    def train_classifier_head(self, 
                              train_anndata: AnnData, 
                              base_model: BaseModelProtocol, 
                              head: BaseTaskModel, 
                              gene_names: str = "index",
                              labels_column_name: str = "cell_type",
                              test_size: float = 0.2,
                              random_state: int = 42) -> Self:
        """Train the classification head. The base model is used to generate the embeddings, which are then used to train the model head.

        Parameters
        ----------
        train_anndata : AnnData
            The data to train the model head on.
        base_model : BaseModelProtocol
            The base model to generate the embeddings.
        head : BaseTaskModel
            The classification model head to train.
         gene_names : str
            The name of the column in the var attribute of the AnnData object that contains the gene names.
            Default is 'index'.
        labels_column_name : str
            The name of the column in the obs attribute of the AnnData object that contains the labels.
        test_size : float
            The size of the test set.
        random_state : int
            The random state for the train/test split.
    
        Raises
        ------
        TypeError
            If the base_model is not an instance of a class implementing 'BaseModelProtocol'.
            

        Returns
        -------
        An instance of the Classifier class, which contains the base_model and the trained model head.
        """
        
        self.base_model = base_model
        self.name = f"{base_model.__class__.__name__} with {head.__class__.__name__}"
        self._check_validity_for_training(train_anndata, labels_column_name, base_model)

        # first, get the embeddings
        LOGGER.info(f"Getting training embeddings with {base_model.__class__.__name__}.")
        dataset = base_model.process_data(train_anndata, gene_names)
        self.gene_names = gene_names
        x = base_model.get_embeddings(dataset)
          
        # then, train the classification model
        LOGGER.info(f"Training classification model '{self.name}'.")
        y = np.array(train_anndata.obs[labels_column_name].tolist())
        num_classes = len(np.unique(y))
        
        X_train, X_test, y_train, y_test = train_test_split(x, y, test_size = test_size, random_state = random_state)
        head.compile(num_classes, x.shape[1])
        self.trained_task_model = head.train(X_train, y_train, validation_data=(X_test, y_test))
       
        return self

    def load_model(self,
                   base_model: Optional[BaseModelProtocol], 
                   classification_model: ClassificationModelProtocol, 
                   name: str,
                   gene_names: str = "index") -> Self:
        """
        Load a classifier model.
        - If no base model is provided, it is assumed that the classification_model can directly classify data.
            This classificaiton_model must follow the ClassificationModelProtocol and implement a predict method.
        - If a base model is provided, the data is processed by the base model and the embeddings are used as input to the classification_model.
            The base_model must follow the BaseModelProtocol and implement a process_data and get_embeddings method.

        Parameters
        ----------
        base_model : None, BaseModelProtocol
            The optional base model to generate the embeddings.
        classification_model : ClassificationModelProtocol
            The classification model to load.
        name : str
            The name of the model.
        gene_names : str
            The name of the default column (specific to this loaded model) in the var attribute of the AnnData object that contains the gene names.
            This can be useful for models like Geneformer that have a default ("ensemble_id") which is different to the other models ("index").
    
        Raises
        ------
        TypeError
            If the base_model is not an instance of a class implementing 'BaseModelProtocol' or 
            the classification_model is not an instance of a class implementing 'ClassificationModelProtocol'.

        Returns
        -------
        An instance of the Classifier class, which contains the (optional) base_model and the trained model head.
        """

        if base_model and not isinstance(base_model, BaseModelProtocol):
            message = "Expected an instance of a class implementing 'BaseModelProtocol'."
            LOGGER.error(message)
            raise TypeError(message)
        
        if not isinstance(classification_model, ClassificationModelProtocol):
            message = "Expected an instance of a class implementing 'ClassificationModelProtocol'."
            LOGGER.error(message)
            raise TypeError(message)
        
        self.base_model = base_model
        self.trained_task_model = classification_model
        self.name = name
        self.gene_names = gene_names
        return self
    
    def _check_validity_for_training(self, train_anndata: AnnData, labels_column_name: str, base_model: BaseModelProtocol) -> None:
        """
        Check if the data and the base model are valid for training the classifier head.

        Parameters
        ----------
        train_anndata : AnnData
            The data to train the model head on.
        labels_column_name : str
            The name of the column in the obs attribute of the AnnData object that contains the labels.
        base_model : None, BaseModelProtocol
            The base model to generate the embeddings.
            
        Raises
        ------
        TypeError
            If the base_model is not an instance of a class implementing 'BaseModelProtocol' or 
            the labels_column_name is not found in the training data.
        """
        error = False

        # first check
        if not isinstance(base_model, BaseModelProtocol):
            message = "To train a classifier head, a base model of type 'BaseModelProtocol' needs to generate the embeddings first."
            error = True
        
        # second check
        try:
            train_anndata.obs[labels_column_name]
        except KeyError:
            message = f"Column {labels_column_name} not found in the evaluation data."
            error = True
        
        # raise error if any of the checks failed
        if error:
            LOGGER.error(message)
            raise TypeError(message)


from helical import Mamba2mRNA, Mamba2mRNAConfig
import hydra
from omegaconf import DictConfig

@hydra.main(version_base=None, config_path="../run_models/configs", config_name="mamba2_mrna_config")
def run(cfg: DictConfig):
    input_sequences = ["ACUG"*20, "AUGC"*20, "AUGC"*20, "ACUG"*20, "AUUG"*20]

    mamba2_mrna_config = Mamba2mRNAConfig(**cfg)
    mamba2_mrna = Mamba2mRNA(mamba2_mrna_config)

    processed_input_data = mamba2_mrna.process_data(input_sequences)

    embeddings = mamba2_mrna.get_embeddings(processed_input_data)
    print(embeddings.shape)

if __name__ == "__main__":
    run()


import json
import os
import torch
from transformers import PreTrainedModel
import re
from .standalone_hyenadna import HyenaDNAModel
import logging 

LOGGER = logging.getLogger(__name__)

# helper 1
def inject_substring(orig_str):
    """Hack to handle matching keys between models trained with and without
    gradient checkpointing."""

    # modify for mixer keys
    pattern = r"\.mixer"
    injection = ".mixer.layer"

    modified_string = re.sub(pattern, injection, orig_str)

    # modify for mlp keys
    pattern = r"\.mlp"
    injection = ".mlp.layer"

    modified_string = re.sub(pattern, injection, modified_string)

    return modified_string    

# helper 2
def load_weights(scratch_dict, pretrained_dict, checkpointing=False):
        """Loads pretrained (backbone only) weights into the scratch state dict."""

        # loop thru state dict of scratch
        # find the corresponding weights in the loaded model, and set it

        # need to do some state dict "surgery"
        for key, value in scratch_dict.items():
            if 'backbone' in key:
                # the state dicts differ by one prefix, '.model', so we add that
                key_loaded = 'model.' + key
                # breakpoint()
                # need to add an extra ".layer" in key
                if checkpointing:
                    key_loaded = inject_substring(key_loaded)
                try:
                    scratch_dict[key] = pretrained_dict[key_loaded]
                except:
                    raise Exception('key mismatch in the state dicts!')

        # scratch_dict has been updated
        return scratch_dict

class HyenaDNAPreTrainedModel(PreTrainedModel):
    """
    An abstract class to handle weights initialization and a simple interface for downloading and loading pretrained
    models.
    """
    base_model_prefix = "hyenadna"

    def __init__(self):
        pass

    def forward(self, input_ids, **kwargs):
        return self.model(input_ids, **kwargs)

    @classmethod
    def from_pretrained(cls,
                        config,
                        device='cpu',
                        use_head=False,
                        n_classes=2,
                      ):
        scratch_model = HyenaDNAModel(**config, use_head=use_head, n_classes=n_classes)  # the new model format
        loaded_ckpt = torch.load(config["model_path"], map_location=torch.device(device))

        # need to load weights slightly different if using gradient checkpointing
        if config.get("checkpoint_mixer", False):
            checkpointing = config["checkpoint_mixer"] == True or config["checkpoint_mixer"] == True
        else:
            checkpointing = False

        # grab state dict from both and load weights
        state_dict = load_weights(scratch_model.state_dict(), loaded_ckpt['state_dict'], checkpointing=checkpointing)

        # scratch model has now been updated
        scratch_model.load_state_dict(state_dict)
        LOGGER.info("Loaded pretrained weights ok!")
        return scratch_model
    


from helical import Mamba2mRNAFineTuningModel, Mamba2mRNAConfig
import hydra
from omegaconf import DictConfig

@hydra.main(version_base=None, config_path="../run_models/configs", config_name="mamba2_mrna_config")
def run_fine_tuning(cfg: DictConfig):
    input_sequences = ["ACUG"*20, "AUGC"*20, "AUGC"*20, "ACUG"*20, "AUUG"*20]
    labels = [0, 2, 2, 0, 1]

    mamba2_mrna_config = Mamba2mRNAConfig(**cfg)
    mamba2_mrna_fine_tune = Mamba2mRNAFineTuningModel(mamba2_mrna_config=mamba2_mrna_config, fine_tuning_head="classification", output_size=3)

    train_dataset = mamba2_mrna_fine_tune.process_data(input_sequences)

    mamba2_mrna_fine_tune.train(train_dataset=train_dataset, train_labels=labels)

    outputs = mamba2_mrna_fine_tune.get_outputs(train_dataset)
    print(outputs.shape)

if __name__ == "__main__":
    run_fine_tuning()



import numpy as np
from datasets import DatasetDict
from helical.models.hyena_dna.model import HyenaDNA
from helical.models.hyena_dna.hyena_dna_config import HyenaDNAConfig    
from datasets import get_dataset_config_names
from datasets import load_dataset
import os
from typing import Tuple
from tqdm import tqdm

configurer = HyenaDNAConfig(model_name="hyenadna-tiny-1k-seqlen-d256")
hyena_model = HyenaDNA(configurer=configurer)

def get_model_inputs(dataset: DatasetDict, nbr_sequences: int = 50) -> Tuple[np.ndarray, np.ndarray]:
    """
    Parameters
    ----------
    dataset : DatasetDict
        The dataset containing the sequences and labels.
    nbr_sequences : int, optional
        The number of sequences to process at a time, by default 50.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        A tuple containing the input features and labels.
    """
    x_tot = np.empty((0, configurer.config['d_model'])) 
    labels_tot = np.empty((0,), dtype=int)
    
    # disable logging to avoid cluttering the output
    import logging
    logging.disable(logging.CRITICAL)

    for i in tqdm(range(0, len(dataset), nbr_sequences)):
        tokenized_sequences = hyena_model.process_data(dataset["sequence"][i:i+nbr_sequences])    
        embeddings = hyena_model.get_embeddings(tokenized_sequences[0])
        numpy_array = embeddings.cpu().detach().numpy()
        x = numpy_array.mean(axis=1)
            
        # normalize the data
        x = (x - np.mean(x, axis=0)) / np.std(x, axis=0)
        x_tot = np.concatenate((x_tot, x), axis=0)

        labels = np.array(dataset["label"][i:i+nbr_sequences])
        labels_tot = np.concatenate((labels_tot, labels), axis=0)
    return x_tot, labels_tot

labels = get_dataset_config_names("InstaDeepAI/nucleotide_transformer_downstream_tasks")

for i, label in enumerate(labels):
    print(f"Processing '{label}' dataset: {i+1} of {len(labels)}")

    dataset = load_dataset("InstaDeepAI/nucleotide_transformer_downstream_tasks", label)
    x, y = get_model_inputs(dataset["train"], 50)

    if not os.path.exists("data"):
        os.makedirs("data/train")
        os.makedirs("data/test")


    np.save(f"data/train/x_{label}_norm_256", x)
    np.save(f"data/train/y_{label}_norm_256", y)

    X_unseen, y_unseen = get_model_inputs(dataset["test"], 50)
    np.save(f"data/test/x_{label}_norm_256", X_unseen)
    np.save(f"data/test/y_{label}_norm_256", y_unseen)



import pytest
import torch
from helical.models.geneformer.model import Geneformer
from helical.models.geneformer.geneformer_config import GeneformerConfig
from helical.models.geneformer.geneformer_utils import get_embs, load_model
from helical.models.geneformer.fine_tuning_model import GeneformerFineTuningModel
from anndata import AnnData

class TestGeneformerModel:
    @pytest.fixture(params=["gf-12L-30M-i2048", "gf-12L-95M-i4096"])
    def geneformer(self, request):
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        config = GeneformerConfig(model_name=request.param, device=self.device)
        return Geneformer(config)

    @pytest.fixture
    def mock_data(self):
        data = AnnData()
        data.var['gene_symbols'] = ['SAMD11', 'PLEKHN1', 'HES4']
        data.var['ensembl_id'] = ['ENSG00000187634', 'ENSG00000187583', 'ENSG00000188290']
        data.obs["cell_type"] = ["CD4 T cells"]
        data.X = [[1, 2, 5]]
        return data

    @pytest.fixture
    def fine_tune_mock_data(self):
        labels = list([0])
        return labels

    def test_pass_invalid_model_name(self):
        with pytest.raises(ValueError):
            geneformer_config = GeneformerConfig(model_name='InvalidName')
        

    def test_process_data_mapping_to_ensemble_ids(self, geneformer, mock_data):
        assert mock_data.var['ensembl_id'][0] == 'ENSG00000187634'
        # is the same as the above line but more verbose (linking the gene symbol to the ensembl id)
        assert mock_data.var[mock_data.var['gene_symbols'] == 'SAMD11']['ensembl_id'].values[0] == 'ENSG00000187634'
        assert mock_data.var[mock_data.var['gene_symbols'] == 'PLEKHN1']['ensembl_id'].values[0] == 'ENSG00000187583'
        assert mock_data.var[mock_data.var['gene_symbols'] == 'HES4']['ensembl_id'].values[0] == 'ENSG00000188290'

    def test_process_data_padding_and_masking_ids(self, geneformer, mock_data):
        # for this token mapping, the padding token is 0 and the mask token is 1
        geneformer.process_data(mock_data, gene_names='gene_symbols')
        assert geneformer.gene_token_dict.get("<pad>") == 0
        assert geneformer.gene_token_dict.get("<mask>") == 1

    def test_ensure_data_validity_raising_error_with_missing_ensembl_id_column(self, geneformer, mock_data):
        del mock_data.var['ensembl_id']
        with pytest.raises(KeyError):
            geneformer.ensure_rna_data_validity(mock_data, "ensembl_id")
    
    @pytest.mark.parametrize("gene_symbols, raises_error",
                             [
                                (['ENSGSAMD11', 'ENSGPLEKHN1', 'ENSGHES4'], True), # humans
                                (['ENSMUSG00000021033', 'ENSMUSG00000021033', 'ENSMUSG00000021033'], True), # mice
                                (['SAMD11', 'None', 'HES4'], True),
                                (['SAMD11', 'PLEKHN1', 'HES4'], False),
                             ]
    )
    def test_ensembl_data_is_caught(self, geneformer, mock_data, gene_symbols, raises_error):
        mock_data.var['gene_symbols'] = gene_symbols
        if raises_error:
            with pytest.raises(ValueError):
                geneformer.process_data(mock_data, "gene_symbols")
        else:
            geneformer.process_data(mock_data, "gene_symbols")

    def test_cls_mode_with_v1_model_config(self, geneformer, mock_data):
        if geneformer.config["special_token"]:
            pytest.skip("This test is only for v1 models and should thus be only executed once.")
        with pytest.raises(ValueError):
            config = GeneformerConfig(model_name="gf-12L-30M-i2048", device="cpu", emb_mode='cls')

    def test_get_embs_cell_mode(self, geneformer, mock_data):
        tokenized_dataset = geneformer.process_data(mock_data, gene_names='gene_symbols')
        model = load_model("Pretrained", geneformer.files_config["model_files_dir"], self.device)
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        embs = get_embs(
            model,
            tokenized_dataset,
            emb_mode="cell",
            layer_to_quant=-1,
            pad_token_id=geneformer.pad_token_id,
            forward_batch_size=1,
            token_gene_dict=geneformer.gene_token_dict,
            device=device
        )
        assert embs.shape == (1, model.config.hidden_size)

    def test_get_embs_cls_mode(self, geneformer, mock_data):
        if not geneformer.config["special_token"]:
            pytest.skip("This test is only for models with special tokens (v2)")
        tokenized_dataset = geneformer.process_data(mock_data, gene_names='gene_symbols')
        model = load_model("Pretrained", geneformer.files_config["model_files_dir"], self.device)
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        embs = get_embs(
            model,
            tokenized_dataset,
            emb_mode="cls",
            layer_to_quant=-1,
            pad_token_id=geneformer.pad_token_id,
            forward_batch_size=1,
            gene_token_dict=geneformer.gene_token_dict,
            device=device
        )
        assert embs.shape == (1, model.config.hidden_size)

    def test_get_embs_gene_mode(self, geneformer, mock_data):
        tokenized_dataset = geneformer.process_data(mock_data, gene_names='gene_symbols')
        model = load_model("Pretrained", geneformer.files_config["model_files_dir"], self.device)
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        embs = get_embs(
            model,
            tokenized_dataset,
            emb_mode="gene",
            layer_to_quant=-1,
            pad_token_id=geneformer.pad_token_id,
            forward_batch_size=1,
            gene_token_dict=geneformer.gene_token_dict,
            device=device
        )
        assert embs.shape[0] == 1
        assert embs.shape[2] == model.config.hidden_size

    def test_get_embs_different_layer(self, geneformer, mock_data):
        tokenized_dataset = geneformer.process_data(mock_data, gene_names='gene_symbols')
        model = load_model("Pretrained", geneformer.files_config["model_files_dir"], self.device)
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        embs_last = get_embs(
            model,
            tokenized_dataset,
            emb_mode="cell",
            layer_to_quant=-1,
            pad_token_id=geneformer.pad_token_id,
            forward_batch_size=1,
            gene_token_dict=geneformer.gene_token_dict,
            device=device
        )
        embs_first = get_embs(
            model,
            tokenized_dataset,
            emb_mode="cell",
            layer_to_quant=0,
            pad_token_id=geneformer.pad_token_id,
            forward_batch_size=1,
            gene_token_dict=geneformer.gene_token_dict,
            device=device
        )
        assert not torch.allclose(embs_last, embs_first)

    def test_get_embs_cell_mode(self, geneformer, mock_data):
        tokenized_dataset = geneformer.process_data(mock_data, gene_names='gene_symbols')
        model = load_model("Pretrained", geneformer.files_config["model_files_dir"], self.device)
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        embs = get_embs(
            model,
            tokenized_dataset,
            emb_mode="cell",
            layer_to_quant=-1,
            pad_token_id=geneformer.pad_token_id,
            forward_batch_size=1,
            gene_token_dict=geneformer.gene_token_dict,
            device=device
        )
        assert embs.shape == (1, model.config.hidden_size)

    def test_cls_eos_tokens_presence(self, geneformer, mock_data):
        geneformer.process_data(mock_data, gene_names='gene_symbols')
        if geneformer.config["special_token"]:
            assert "<cls>" in geneformer.tk.gene_token_dict
            assert "<eos>" in geneformer.tk.gene_token_dict
        else:
            assert "<cls>" not in geneformer.tk.gene_token_dict
            assert "<eos>" not in geneformer.tk.gene_token_dict

    def test_model_input_size(self, geneformer):
        assert geneformer.config["input_size"] == geneformer.configurer.model_map[geneformer.config["model_name"]]['input_size']

    def test_fine_tune_classifier_returns_correct_shape(self, mock_data, fine_tune_mock_data):
        device = "cuda" if torch.cuda.is_available() else "cpu"
        fine_tuned_model = GeneformerFineTuningModel(GeneformerConfig(device=device), fine_tuning_head="classification", output_size=1)
        tokenized_dataset = fine_tuned_model.process_data(mock_data, gene_names='gene_symbols')
        tokenized_dataset = tokenized_dataset.add_column('labels', fine_tune_mock_data)
        
        fine_tuned_model.train(train_dataset=tokenized_dataset, label='labels')
        assert fine_tuned_model is not None
        outputs = fine_tuned_model.get_outputs(tokenized_dataset)
        assert outputs.shape == (len(mock_data), len(fine_tune_mock_data))

            


from helical import UCEConfig, UCEFineTuningModel
from helical.utils import get_anndata_from_hf_dataset
from datasets import load_dataset
from omegaconf import DictConfig
import hydra

@hydra.main(version_base=None, config_path="../run_models/configs", config_name="uce_config")
def run_fine_tuning(cfg: DictConfig):
    hf_dataset = load_dataset("helical-ai/yolksac_human",split="train[:5%]", trust_remote_code=True, download_mode="reuse_cache_if_exists")
    ann_data = get_anndata_from_hf_dataset(hf_dataset)

    cell_types = ann_data.obs["LVL1"][:10].tolist()

    label_set = set(cell_types)

    uce_config=UCEConfig(**cfg)
    uce_fine_tune = UCEFineTuningModel(uce_config=uce_config, fine_tuning_head="classification", output_size=len(label_set))

    dataset = uce_fine_tune.process_data(ann_data[:10], name="train")

    class_id_dict = {label: i for i, label in enumerate(label_set)}
    cell_types = [class_id_dict[cell] for cell in cell_types]

    uce_fine_tune.train(train_input_data=dataset, train_labels=cell_types)
    
if __name__ == "__main__":
    run_fine_tuning()


from helical.models.classification.classifier import Classifier
from anndata import AnnData
import logging
from numpy import ndarray
import numpy as np
from sklearn.metrics import accuracy_score, precision_score, f1_score, recall_score
from scib.metrics import metrics
from omegaconf import DictConfig
from helical.models.base_models import BaseModelProtocol

LOGGER = logging.getLogger(__name__)

def evaluate_integration(model_list: list[tuple[BaseModelProtocol, str]], adata: AnnData, data_cfg: DictConfig, integration_cfg: DictConfig) -> dict[str, dict[str, float]]:
    """
    Evaluate the data integration of the anndata object using the scib metrics. 

    Parameters
    ----------
    model_list : list[tuple[BaseModelProtocol, str]]
        A list of tuples containing:
        A model adhering to the BaseModelProtocol that is used to generate the embeddings.
        The name of that model.
    adata : AnnData
        The AnnData object that contains the embeddings.
    data_cfg : DictConfig
        The configuration of the data. It must contain the keys "batch_key" and "label_key".
    integration_cfg : DictConfig
        The configuration of the integration.

    Raises
    ------
    TypeError
        If the model is not an instance of BaseModelProtocol.
        
    Returns
    -------
    A dictionary containing the evaluations for each model specified in the tuple.

    """
    evaluations = {}
    for model, model_name in model_list:
        LOGGER.info(f"Processing integration evaluation using: {model_name}")
        embed_obsm_name = f"X_{model_name}"

        if not isinstance(model, BaseModelProtocol):
            message = "To train a classifier head, a base model of type 'BaseModelProtocol' needs to generate the embeddings first."
            LOGGER.error(message)
            raise TypeError(message)
        
        # because some models may process the data in different ways (including scib) we use .copy() for each model
        # so that all models start with the same data and are independent of each other
        # otherwise, some evaluations will be identical and thus incorrect 
        adata_copy = adata.copy()

        dataset = model.process_data(adata_copy, gene_names=data_cfg["gene_names"])
        adata_copy.obsm[embed_obsm_name] = model.get_embeddings(dataset)
        evaluation = _get_integration_evaluations(adata_copy,
                                                  adata_copy,
                                                  data_cfg["batch_key"], 
                                                  data_cfg["label_key"], 
                                                  embed_obsm_name, 
                                                  **integration_cfg)
        evaluations.update({model_name: evaluation})
    return evaluations

def evaluate_classification(models: list[Classifier], eval_anndata: AnnData, labels_column_name: str) -> dict[str, dict[str, float]]:
    """
    Evaluate the classification models using the evaluation dataset. 
    The evaluation dataset is used to calculate the accuracy, precision, f1, and recall scores.

    Parameters
    ----------
    models : list[Classifier]
        The list of models to evaluate.
    eval_anndata : AnnData
        The evaluation data.
    labels_column_name : str
        The name of the column in the obs attribute of the AnnData object that contains the labels.

    Raises
    ------
    TypeError
        If the column with the labels is not found in the evaluation data.

    Returns
    -------
    A dictionary containing the evaluations for each HelicalBaseFoundationModel provided in the initialization.

    """
    try:
        eval_labels = np.array(eval_anndata.obs[labels_column_name].tolist())
    except KeyError:
        message = f"Column {labels_column_name} not found in the evaluation data."
        LOGGER.error(message)
        raise TypeError(message)

    eval_anndata = eval_anndata
    evaluations = {}
    for model in models:
        LOGGER.info(f"Processing classification evaluation embeddings using {model.name}.")
        prediction = model.get_predictions(eval_anndata)
        evaluation = _get_classification_evaluations(prediction, eval_labels)
        evaluations.update({model.name: evaluation})
    return evaluations

def _get_integration_evaluations(adata: AnnData, 
                                 adata_int: AnnData,
                                 batch_key: str, 
                                 label_key: str, 
                                 embed_obsm_name: str, 
                                 **configs) -> dict[str, float]:
    """
    Wrapper for all metrics used in the scib study.
    Metrics are computed according to the provided configs.
    Have a look at the User Guide for more information: 
    https://scib.readthedocs.io/en/latest/user_guide.html
    
    Parameters
    ----------
    adata : AnnData
        Unintegrated, preprocessed anndata object
    adata_int : AnnData
        Integrated anndata object
    batch_key : str
        Name of the batch column in adata.obs and adata_int.obs
    label_key : str
        Name of the biological label (cell type) column in adata.obs and adata_int.obs
    embed_obsm_name : str
        The embedding representation in adata_int.obsm
        Used for:
            + silhouette scores (label ASW, batch ASW),
            + PC regression,
            + cell cycle conservation,
            + isolated label scores, and
            + kBET
    configs : DictConfig
        Configuration for the metrics calculation as key-value pairs, coming from the config file.
    """

    results = metrics(adata, adata_int, batch_key, label_key, embed = embed_obsm_name, **configs["scib"])
    result_dict = results[0].to_dict()

    # rearrange 
    evaluations = {
                    "BIO": 
                        {
                            "ARI_cell": result_dict["ARI_cluster/label"],
                            "NMI_cell": result_dict["NMI_cluster/label"],
                            "ASW_cell": result_dict["ASW_label"],
                            },
                    "BATCH": 
                        {
                            "ASW_batch": result_dict["ASW_label/batch"],
                            "Graph_Conn": result_dict["graph_conn"],
                            },
    }
    return evaluations


def _get_classification_evaluations(y_true: ndarray, y_pred: ndarray) -> dict[str, float]:
    """
    Based on the predictions and the ground truth, calculate evaluation metrics: accuracy, precision, f1, recall.

    Parameters
    ----------
    y_pred : ndarray
        The predicted labels.
    
    y_true : ndarray
        The ground truth labels.

    Returns
    -------
    A dictionary containing the evaluations.
    """
    evaluation = {
        "Accuracy": accuracy_score(y_true, y_pred),
        "Precision": precision_score(y_true, y_pred, average='macro'),
        "F1": f1_score(y_true, y_pred, average='macro'),
        "Recall": recall_score(y_true, y_pred, average='macro'),
    }        
    return evaluation


from helical import HelixmRNAFineTuningModel, HelixmRNAConfig
import hydra
from omegaconf import DictConfig

@hydra.main(version_base=None, config_path="../run_models/configs", config_name="helix_mrna_config")
def run_fine_tuning(cfg: DictConfig):
    input_sequences = ["EACU"*20, "EAUG"*20, "EAUG"*20, "EACU"*20, "EAUU"*20]
    labels = [0, 2, 2, 0, 1]

    helix_mrna_config = HelixmRNAConfig(**cfg)
    helix_mrna_fine_tune = HelixmRNAFineTuningModel(helix_mrna_config=helix_mrna_config, fine_tuning_head="classification", output_size=3)

    train_dataset = helix_mrna_fine_tune.process_data(input_sequences)

    helix_mrna_fine_tune.train(train_dataset=train_dataset, train_labels=labels)

    outputs = helix_mrna_fine_tune.get_outputs(train_dataset)
    print(outputs.shape)

if __name__ == "__main__":
    run_fine_tuning()



from helical.models.scgpt.model import scGPT, scGPTConfig
from helical.models.scgpt.fine_tuning_model import scGPTFineTuningModel
from anndata import AnnData
from helical.models.scgpt.tokenizer import GeneVocab
import pytest
import anndata as ad
import numpy as np
from scipy.sparse import csr_matrix

class TestSCGPTModel:
    scgpt = scGPT()

    # Create a dummy AnnData object
    data = AnnData()
    data.var["gene_names"] = ['SAMD11', 'PLEKHN1', "NOT_IN_VOCAB", "<pad>", 'HES4']
    data.obs["cell_type"] = ["CD4 T cells"]
    
    vocab = {
        "SAMD11": 0,
        "PLEKHN1": 1,
        "HES4": 2,
        "<pad>": 3,
    }
    scgpt.vocab = GeneVocab.from_dict(vocab)
    
    data.X = [[1, 2, 5, 6, 0]]
    
    def test_process_data(self):
        dataset = self.scgpt.process_data(self.data, gene_names = "gene_names")

        assert self.scgpt.gene_names == "gene_names"
        assert self.scgpt.gene_names in self.data.var

        # asserts that all the genes in gene_names have been correctly translated 
        # to the corresponding ids based on the vocabulary
        assert (self.data.var["id_in_vocab"] == [0, 1, -1, 3, 2]).all()

        # make sure that the genes not present in the vocabulary are filtered out
        # meaning -1 is not present in the gene_ids
        assert (dataset.gene_ids == [0, 1, 3, 2]).all()

        # ensure that the default index of the vocabulary is set to the id of the pad token
        assert self.scgpt.vocab.get_default_index() == 3

        assert (dataset.count_matrix == [1, 2, 6, 0]).all()

    def test_correct_handling_of_batch_ids(self):
        batch_id_array = [1]
        self.data.obs["batch_id"] = batch_id_array
        dataset = self.scgpt.process_data(self.data, gene_names = "gene_names", use_batch_labels=True)
        assert (dataset.batch_ids == batch_id_array).all()

    def test_direct_assignment_of_genes_to_index(self):
        self.data.var.index = ['SAMD11', 'PLEKHN1', "NOT_IN_VOCAB", "<pad>", 'HES4']
        self.scgpt.process_data(self.data, gene_names = "index")
        
        # as set above, the gene column can also be direclty assigned to the index column
        assert self.scgpt.gene_names == "index"
        assert self.scgpt.gene_names in self.data.var


    def test_get_embeddings(self):
        dataset = self.scgpt.process_data(self.data, gene_names = "gene_names")
        embeddings = self.scgpt.get_embeddings(dataset)
        assert embeddings.shape == (1, 512)

    dummy_data = ad.read_h5ad("ci/tests/data/cell_type_sample.h5ad")
    @pytest.mark.parametrize("data, gene_names, batch_labels", 
                             [
                                #  missing gene_names in data.var
                                (AnnData(), "gene_names", False),
                                #  missing batch_id in data.obs
                                (dummy_data, "index", True),
                             ]
    )
    def test_ensure_data_validity__key_error(self, data, gene_names, batch_labels):
        with pytest.raises(KeyError):
            self.scgpt.ensure_data_validity(data, gene_names, batch_labels)
    
    err_np_arr_data = ad.read_h5ad("ci/tests/data/cell_type_sample.h5ad")
    err_np_arr_data.X.dtype=float
    err_np_arr_data.X[0,0] = 0.5

    err_csr_data = ad.read_h5ad("ci/tests/data/cell_type_sample.h5ad")
    err_csr_data.X = csr_matrix(np.random.rand(100, 5), dtype=np.float32)
    @pytest.mark.parametrize("data",
                             [
                                (err_np_arr_data),
                                (err_csr_data),
                             ]
    )
    def test_ensure_data_validity__value_error(self, data):
        '''The data in X must be ints. Test an error is raised for both np.ndarray and csr_matrix.'''
        with pytest.raises(ValueError):
            self.scgpt.ensure_data_validity(data, "index", False)
        assert "total_counts" in data.obs

    np_arr_data = ad.read_h5ad("ci/tests/data/cell_type_sample.h5ad")
    csr_data = ad.read_h5ad("ci/tests/data/cell_type_sample.h5ad")
    csr_data.X = csr_matrix(np.random.poisson(1, size=(100, 5)), dtype=np.float32)
    @pytest.mark.parametrize("data",
                             [
                                (np_arr_data),
                                (csr_data),
                             ]
    )
    def test_ensure_data_validity__no_error(self, data):
        '''The data in X must be ints. Test no error is raised for both np.ndarray and csr_matrix.'''
        self.scgpt.ensure_data_validity(data, "index", False)
        assert "total_counts" in data.obs

    def test_fine_tune_classification_returns_correct_shape(self):
        labels = list([0])
        fine_tuned_model = scGPTFineTuningModel(scGPTConfig(), fine_tuning_head="classification", output_size=1)
        tokenized_dataset = fine_tuned_model.process_data(self.data)
        fine_tuned_model.train(train_input_data=tokenized_dataset, train_labels=labels)
        assert fine_tuned_model is not None
        outputs = fine_tuned_model.get_outputs(tokenized_dataset)
        assert outputs.shape == (len(self.data), len(labels))


from helical.models.hyena_dna.model import HyenaDNA, HyenaDNAConfig
import hydra
from omegaconf import DictConfig

@hydra.main(version_base=None, config_path="configs", config_name="hyena_dna_config")
def run(cfg: DictConfig):

    hyena_config = HyenaDNAConfig(**cfg)
    model = HyenaDNA(configurer = hyena_config)   
    sequence = 'ACTG' * int(1024/4)
    tokenized_sequence = model.process_data(sequence)
    embeddings = model.get_embeddings(tokenized_sequence)
    print(embeddings.shape)

if __name__ == "__main__":
    run()


from enum import Enum

class LoggingType(str, Enum):
    FILE='FILE'
    CONSOLE='CONSOLE'
    FILE_AND_CONSOLE='FILE_AND_CONSOLE'
    NOTSET='NOTSET'

class LoggingLevel(str, Enum):
    CRITICAL = "CRITICAL"
    FATAL = "FATAL"
    ERROR = "ERROR"
    WARNING = "WARNING"
    WARN = "WARN"
    INFO = "INFO"
    DEBUG = "DEBUG"
    NOTSET = "NOTSET"


from helical.utils.downloader import Downloader
from pathlib import Path
import logging
LOGGER = logging.getLogger(__name__)

def download_geneformer_models():
    downloader = Downloader()
    versions = ['v1', 'v2']

    # We can decide to download more models by simply adding the model names from the full list as reported in geneformer_config.py
    version_models_dict = {
        "v1": ["gf-12L-30M-i2048"],
        "v2": ["gf-12L-95M-i4096"]
    }
    
    for version in versions:
        # Download common files for each version
        common_files = [
            f"geneformer/{version}/gene_median_dictionary.pkl",
            f"geneformer/{version}/token_dictionary.pkl",
            f"geneformer/{version}/ensembl_mapping_dict.pkl",
        ]
        for file in common_files:
            downloader.download_via_name(file)
        
        # Get all model directories
        model_dirs = version_models_dict[version]
        
        for model_name in model_dirs:
            model_files = [
                f"geneformer/{version}/{model_name}/config.json",
                f"geneformer/{version}/{model_name}/training_args.bin",
            ]
            
            # Add version-specific files
            if version == 'v2':
                model_files.extend([
                    f"geneformer/{version}/{model_name}/generation_config.json",
                    f"geneformer/{version}/{model_name}/model.safetensors",
                ])
            else:
                model_files.append(f"geneformer/{version}/{model_name}/pytorch_model.bin")
            
            # Download all files for the current model
            for file in model_files:
                downloader.download_via_name(file)
    
    LOGGER.info("All Geneformer models and files have been downloaded.")

def main():
    downloader = Downloader()
    downloader.display = False

    downloader.download_via_name("uce/4layer_model.torch")
    downloader.download_via_name("uce/all_tokens.torch")
    downloader.download_via_name("uce/species_chrom.csv")
    downloader.download_via_name("uce/species_offsets.pkl")
    downloader.download_via_name("uce/protein_embeddings/Homo_sapiens.GRCh38.gene_symbol_to_embedding_ESM2.pt")
    downloader.download_via_name("uce/protein_embeddings/Macaca_fascicularis.Macaca_fascicularis_6.0.gene_symbol_to_embedding_ESM2.pt")
    
    downloader.download_via_name("scgpt/scGPT_CP/vocab.json")
    downloader.download_via_name("scgpt/scGPT_CP/best_model.pt")

    download_geneformer_models()

    downloader.download_via_name("hyena_dna/hyenadna-tiny-1k-seqlen.ckpt")
    downloader.download_via_name("hyena_dna/hyenadna-tiny-1k-seqlen-d256.ckpt")

    downloader.download_via_link(Path("yolksac_human.h5ad"), "https://huggingface.co/datasets/helical-ai/yolksac_human/resolve/main/data/17_04_24_YolkSacRaw_F158_WE_annots.h5ad?download=true")
    return True

if __name__ == "__main__":
    main()



from helical.models.base_models import HelicalRNAModel
import logging
from pathlib import Path
import numpy as np
from anndata import AnnData
from helical.utils.downloader import Downloader
import pickle
from transformers import BertForMaskedLM
from helical.models.geneformer.geneformer_utils import get_embs,quant_layers
from helical.models.geneformer.geneformer_tokenizer import TranscriptomeTokenizer
from helical.models.geneformer.geneformer_config import GeneformerConfig
from helical.utils.mapping import map_gene_symbols_to_ensembl_ids
from datasets import Dataset
from typing import Optional
from accelerate import Accelerator

LOGGER = logging.getLogger(__name__)
class Geneformer(HelicalRNAModel):
    """Geneformer Model. 
    The Geneformer Model is a transformer-based model that can be used to extract gene embeddings from single-cell RNA-seq data. Both versions are made available through this interface.
    Both versions of Geneformer (v1 and v2) have different sub-models with varying numbers of layers, context size and pretraining set. The available models are the following:

    Version 1.0:
    - gf-12L-30M-i2048
    - gf-6L-30M-i2048

    Version 2.0:
    - gf-12L-95M-i4096
    - gf-12L-95M-i4096-CLcancer
    - gf-20L-95M-i4096

    For a detailed explanation of the differences between these models and versions, please refer to the Geneformer model card: https://helical.readthedocs.io/en/latest/docs/Geneformer.html

    Example
    -------
    ```python
     from helical.models import Geneformer, GeneformerConfig
     import anndata as ad
     
     # For Version 2.0
     geneformer_config_v2 = GeneformerConfig(model_name="gf-12L-95M-i4096", batch_size=10)
     geneformer_v2 = Geneformer(configurer=geneformer_config_v2)
    
     # You can use other model names in the config, such as:
     # "gf-12L-30M-i2048" (Version 1.0)
     # "gf-12L-95M-i4096-CLcancer" (Version 2.0, Cancer-tuned)
     # "gf-20L-95M-i4096" (Version 2.0, 20-layer model)
     
     # Example usage for base pretrained model (for general transcriptomic analysis, v1 and v2)
     ann_data = ad.read_h5ad("general_dataset.h5ad")
     dataset = geneformer_v2.process_data(ann_data)
     embeddings = geneformer_v2.get_embeddings(dataset)
     print("Base model embeddings shape:", embeddings.shape)
    
     # Example usage for cancer-tuned model (for cancer-specific analysis)
     cancer_ann_data = ad.read_h5ad("cancer_dataset.h5ad")
     cancer_dataset = geneformer_v2_cancer.process_data(cancer_ann_data)
     cancer_embeddings = geneformer_v2_cancer.get_embeddings(cancer_dataset)
     print("Cancer-tuned model embeddings shape:", cancer_embeddings.shape)
    ```

    Parameters
    ----------
    configurer : GeneformerConfig, optional, default = default_configurer
        The model configration

    Notes
    -----
    The first version of the model is published in this <a href="https://www.nature.com/articles/s41586-023-06139-9.epdf?sharing_token=u_5LUGVkd3A8zR-f73lU59RgN0jAjWel9jnR3ZoTv0N2UB4yyXENUK50s6uqjXH69sDxh4Z3J4plYCKlVME-W2WSuRiS96vx6t5ex2-krVDS46JkoVvAvJyWtYXIyj74pDWn_DutZq1oAlDaxfvBpUfSKDdBPJ8SKlTId8uT47M%3D">Nature Paper</a>. 
    The second version of the model is available at <a href="https://pubmed.ncbi.nlm.nih.gov/39229018/">NIH</a>.
    We use the implementation from the <a href="https://huggingface.co/ctheodoris/Geneformer/tree/main">Geneformer</a> repository.

    """
    default_configurer = GeneformerConfig()
    def __init__(self, configurer: GeneformerConfig = default_configurer) -> None:
        super().__init__()
        self.configurer = configurer
        self.config = configurer.config
        self.files_config = configurer.files_config
        self.device = self.config['device']

        downloader = Downloader()
        for file in self.configurer.list_of_files_to_download:
            downloader.download_via_name(file)

        self.model =  BertForMaskedLM.from_pretrained(self.files_config['model_files_dir'], output_hidden_states=True, output_attentions=False)
        self.model.eval()
        self.model = self.model.to(self.device)


        self.layer_to_quant = quant_layers(self.model) + self.config['emb_layer']
        self.emb_mode = self.config['emb_mode']
        self.forward_batch_size = self.config['batch_size']
        
        if self.config['accelerator']:
            self.accelerator = Accelerator(project_dir=self.configurer.model_dir)
            self.model = self.accelerator.prepare(self.model)
        else:
            self.accelerator = None

        # load token dictionary (Ensembl IDs:token)
        with open(self.files_config["token_path"], "rb") as f:
            self.gene_token_dict = pickle.load(f)
            self.pad_token_id = self.gene_token_dict.get("<pad>")

        self.tk = TranscriptomeTokenizer(custom_attr_name_dict=self.config["custom_attr_name_dict"],
                                         nproc=self.config['nproc'], 
                                         model_input_size=self.config["input_size"],
                                         special_token=self.config["special_token"],
                                         gene_median_file = self.files_config["gene_median_path"], 
                                         token_dictionary_file = self.files_config["token_path"],
                                         gene_mapping_file = self.files_config["ensembl_dict_path"],
                                        )
        
        LOGGER.info(f"Model finished initializing.")
        
    def process_data(self, 
                     adata: AnnData,  
                     gene_names: str = "index", 
                     output_path: Optional[str] = None,
                     use_raw_counts: bool = True,
                     ) -> Dataset:   
        """
        Processes the data for the Geneformer model.

        Parameters
        ----------
        adata : AnnData
            The AnnData object containing the data to be processed. Geneformer uses Ensembl IDs to identify genes 
            and currently supports only human genes. If the AnnData object already has an 'ensembl_id' column, 
            the mapping step can be skipped.
        gene_names : str, optional, default="index"
            The column in `adata.var` that contains the gene names. If set to a value other than "ensembl_id", 
            the gene symbols in that column will be mapped to Ensembl IDs using the 'pyensembl' package, 
            which retrieves mappings from the Ensembl FTP server and loads them into a local database.
            - If set to "index", the index of the AnnData object will be used and mapped to Ensembl IDs.
            - If set to "ensembl_id", no mapping will occur.
            Special case:
                If the index of `adata` already contains Ensembl IDs, setting this to "index" will result in 
                invalid mappings. In such cases, create a new column containing Ensembl IDs and pass "ensembl_id" 
                as the value of `gene_names`.
        output_path : str, optional, default=None
            If specified, saves the tokenized dataset to the given output path.
        use_raw_counts : bool, optional, default=True
            Determines whether raw counts should be used.

        Returns
        -------
        Dataset
            The tokenized dataset in the form of a Huggingface Dataset object.
        """

        self.ensure_rna_data_validity(adata, gene_names, use_raw_counts)

        # map gene symbols to ensemble ids if provided
        if gene_names != "ensembl_id":
            if (adata.var[gene_names].str.startswith("ENS").all()) or (adata.var[gene_names].str.startswith("None").any()):
                message = "It seems an anndata with 'ensemble ids' and/or 'None' was passed. " \
                "Please set gene_names='ensembl_id' and remove 'None's to skip mapping."
                LOGGER.info(message)
                raise ValueError(message)
            adata = map_gene_symbols_to_ensembl_ids(adata, gene_names)

        
        tokenized_cells, cell_metadata = self.tk.tokenize_anndata(adata)

        # tokenized_cells, cell_metadata =  self.tk.tokenize_anndata(adata)
        tokenized_dataset = self.tk.create_dataset(tokenized_cells, cell_metadata, use_generator=False)
        
        if output_path:
            output_path = Path(output_path).with_suffix(".dataset")
            tokenized_dataset.save_to_disk(output_path)
        return tokenized_dataset

    def get_embeddings(self, dataset: Dataset) -> np.array:
        """Gets the gene embeddings from the Geneformer model   

        Parameters
        ----------
        dataset : Dataset
            The tokenized dataset containing the processed data

        Returns
        -------
        np.array
            The gene embeddings in the form of a numpy array
        """
        LOGGER.info(f"Inference started:")
        embeddings = get_embs(
            self.model,
            dataset,
            self.emb_mode,
            self.layer_to_quant,
            self.pad_token_id,
            self.forward_batch_size,
            self.gene_token_dict,
            self.device
        ).cpu().detach().numpy()

        return embeddings



from pathlib import Path
from helical.models.scgpt.tokenizer import GeneVocab, get_default_gene_vocab


def test_gene_vocab():
    gene_vocab = GeneVocab([])
    assert len(gene_vocab) == 0

    gene_vocab.append_token("abc")
    assert len(gene_vocab) == 1

    gene_vocab = GeneVocab(gene_vocab)
    assert gene_vocab["abc"] == 0

    gene_vocab = GeneVocab(["def", "g", "h"], specials=["a", "b", "c"])
    assert gene_vocab["a"] == 0
    assert gene_vocab["def"] == 3
    assert len(gene_vocab) == 6

    gene_vocab = GeneVocab(["a"], specials=["<pad>"], special_first=False)
    assert gene_vocab["<pad>"] == 1
    assert gene_vocab.get_default_index() == 1


def test_gene_vocab_from_dict():
    gene_vocab = GeneVocab.from_dict({"a": 0, "b": 1, "c": 2})
    assert len(gene_vocab) == 3
    assert gene_vocab["a"] == 0
    assert gene_vocab["c"] == 2


def test_gene_vocab_from_file():
    # current file path
    test_file = Path(__file__).parent / "vocab.json"
    gene_vocab = GeneVocab.from_file(test_file)
    assert len(gene_vocab) == 3


def test_gene_vocab_pad_token():
    gene_vocab = GeneVocab(["a", "b", "c"], specials=["<pad>"])
    assert gene_vocab.pad_token is None
    gene_vocab.pad_token = "<pad>"
    assert gene_vocab.pad_token == "<pad>"


def test_get_default_gene_vocab():
    gene_vocab = get_default_gene_vocab()
    assert gene_vocab["A12M1"] == 0



from torch.utils.data import Dataset

class HyenaDNADataset(Dataset):
    """HyenaDNA dataset.

    Parameters
    ----------
        sequences: list
            The list of sequences.
        labels: list, default = None
            The list of labels.
    """
    def __init__(self, sequences, labels=None):
        self.sequences = sequences
        self.labels = labels

    def __len__(self):
        return len(self.sequences)

    def __getitem__(self, idx):
        if self.labels is None:
            return self.sequences[idx]
        else:
            return self.sequences[idx], self.labels[idx]

    def set_labels(self, labels):
        self.labels = labels


from pathlib import Path
import tempfile

import numpy as np
import pytest
from scipy.sparse import csr_matrix
from datasets import Dataset
from anndata import AnnData
import os
import shutil

from helical.models.scgpt.tokenizer import GeneVocab
from helical.models.scgpt.scbank import DataBank, DataTable, MetaInfo, Setting

tmp_dir = tempfile.gettempdir()
dir_path = os.path.dirname(os.path.realpath(__file__))

save_path = Path(dir_path) / "test_output"
save_path.mkdir(parents=True, exist_ok=True)


def clear_files(directory: Path):
    """Helper function to clear files in a dir by removing and creating it again."""
    shutil.rmtree(directory)
    directory.mkdir(parents=True, exist_ok=True)


def test_empty_databank():
    db = DataBank()
    assert db.data_tables == {}
    assert db.settings == Setting()
    assert db.gene_vocab == None

    db = DataBank(meta_info=MetaInfo())
    assert db.data_tables == {}
    assert db.gene_vocab == None

    db = DataBank(
        meta_info=MetaInfo(on_disk_path=save_path),
        settings=Setting(immediate_save=True),
    )
    assert (save_path / "studytable.json").is_file()
    assert (save_path / "manifest.json").is_file()
    clear_files(save_path)


def test_empty_datatable():
    dt = DataTable(name="test")
    assert dt.name == "test"
    assert dt.data is None
    assert not dt.is_loaded


def test_empty_metainfo():
    mi = MetaInfo()
    assert mi.study_ids is None
    assert mi.cell_ids is None


def test_save_load_metainfo():
    mi = MetaInfo(save_path)
    mi.save()
    assert (save_path / "studytable.json").is_file()
    assert (save_path / "manifest.json").is_file()
    assert MetaInfo.from_path(save_path) == mi

    clear_files(save_path)


def test_datatable_save():
    dt = DataTable(name="test")

    file_path = save_path / "test.json"
    # make sure the path does not exist originally
    assert not file_path.exists()
    file_path.parent.mkdir(parents=True, exist_ok=True)

    # catch the exception if the data is not loaded
    with pytest.raises(ValueError):
        dt.save(file_path)

    # actually load some example data and test saving
    dt.data = Dataset.from_dict({"a": [1]})
    dt.save(file_path)
    assert file_path.is_file()

    # delete the file
    file_path.unlink()
    assert not file_path.exists()


def test_meta_info_on_disk_path():
    mi = MetaInfo(on_disk_path=tmp_dir)
    assert mi.on_disk_path == Path(tmp_dir)
    assert mi.on_disk_format == "json"


def test_add_gene_vocab():
    db = DataBank()
    db.gene_vocab = GeneVocab.from_dict({"a": 0, "b": 1, "c": 2})
    assert len(db.gene_vocab) == 3
    assert db.gene_vocab["a"] == 0
    assert "c" in db.gene_vocab

    with pytest.raises(ValueError):
        db.gene_vocab = ["a", "b", "c"]


def test_databank_tokenize():
    indptr = np.array([0, 2, 3, 6])
    indices = np.array([0, 2, 2, 0, 1, 2])
    data = np.array([1, 2, 3, 4, 5, 6])
    data = csr_matrix((data, indices, indptr), shape=(3, 3))
    # data is like:
    # [[1, 0, 2],
    #  [0, 0, 3],
    #  [4, 5, 6]]

    ind2ind = {0: 4, 2: 6}

    tokenized = DataBank()._tokenize(data, ind2ind)
    # tokenized is like:
    # {'id': [0, 1, 2],
    # 'genes': [[4, 6], [6], [4, 6]],
    # 'expressions': [[1, 2], [3], [4, 6]]}

    assert tokenized["id"] == [0, 1, 2]
    assert [d.tolist() for d in tokenized["genes"]] == [[4, 6], [6], [4, 6]]
    assert [d.tolist() for d in tokenized["expressions"]] == [[1, 2], [3], [4, 6]]

    # tokenize numpy array
    data = data.toarray()
    tokenized = DataBank()._tokenize(data, ind2ind)
    assert tokenized["id"] == [0, 1, 2]
    assert [d.tolist() for d in tokenized["genes"]] == [[4, 6], [6], [4, 6]]
    assert [d.tolist() for d in tokenized["expressions"]] == [[1, 2], [3], [4, 6]]

    # test array with rows and cols of only zeros
    data[:, 2] = 0
    tokenized = DataBank()._tokenize(data, ind2ind)
    assert tokenized["id"] == [0, 1]
    assert [d.tolist() for d in tokenized["genes"]] == [[4], [4]]
    assert [d.tolist() for d in tokenized["expressions"]] == [[1], [4]]

    # test array w/ rare non-zero values (_tokenize will auto convert it to sparse)
    data = np.zeros((3, 3))
    data[0, 0] = 1.0
    tokenized = DataBank()._tokenize(data, ind2ind)
    assert tokenized["id"] == [0]
    assert [d.tolist() for d in tokenized["genes"]] == [[4]]
    assert [d.tolist() for d in tokenized["expressions"]] == [[1.0]]

    # add the test for recursive batch calls


def test_databank_load_anndata():
    adata = AnnData(
        X=np.array([[1.0, 2.0, 3.0], [4.0, 0.0, 6.0]]),
        obs={"cell": ["cell1", "cell2"], "study": ["study1", "study2"]},
        var={"gene": ["gene_a", "gene_b", "gene_c"]},
    )
    gene_vocab = {"gene_a": 1, "gene_b": 0, "gene_c": 2}

    # test factory initialization from data
    db = DataBank.from_anndata(
        adata,
        gene_vocab,
        to=save_path,
        main_table_key="X",
        token_col="gene",
    )
    assert db.main_table_key == "X"

    converted_dataset = db.data_tables["X"].data
    assert converted_dataset["id"] == [0, 1]
    assert converted_dataset["genes"] == [[1, 0, 2], [1, 2]]
    assert converted_dataset["expressions"] == [[1.0, 2.0, 3.0], [4.0, 6.0]]

    assert (save_path / "X.datatable.json").is_file()
    assert (save_path / "studytable.json").is_file()
    assert (save_path / "manifest.json").is_file()
    assert (save_path / "gene_vocab.json").is_file()

    # test factory initialization from path
    db = DataBank.from_path(save_path)
    assert db.main_table_key == "X"

    main_dataset = db.data_tables["X"].data
    assert main_dataset["id"] == [0, 1]
    assert main_dataset["genes"] == [[1, 0, 2], [1, 2]]
    assert main_dataset["expressions"] == [[1.0, 2.0, 3.0], [4.0, 6.0]]

    clear_files(save_path)

    # test loading from anndata
    db = DataBank(
        meta_info=MetaInfo(on_disk_format=save_path),
        gene_vocab=GeneVocab.from_dict(gene_vocab),
    )
    data_tables = db.load_anndata(adata, data_keys=["X"], token_col="gene")
    assert len(data_tables) == 1

    converted_dataset = data_tables[0].data
    assert converted_dataset["id"] == [0, 1]
    assert converted_dataset["genes"] == [[1, 0, 2], [1, 2]]
    assert converted_dataset["expressions"] == [[1.0, 2.0, 3.0], [4.0, 6.0]]


def test_databank_load_multiple_anndata_layers():
    adata = AnnData(
        X=np.array([[1.0, 2.0, 3.0], [4.0, 0.0, 6.0]]),
        obs={"cell": ["cell1", "cell2"], "study": ["study1", "study2"]},
        var={"gene": ["gene_a", "gene_b", "gene_c"]},
        layers={"layer1": np.array([[1.0, 2.0, 3.0], [4.0, 0.0, 6.0]])},
    )
    gene_vocab = {"gene_a": 1, "gene_b": 0, "gene_c": 2}

    db = DataBank(meta_info=MetaInfo(), gene_vocab=GeneVocab.from_dict(gene_vocab))
    data_tables = db.load_anndata(adata, token_col="gene")
    assert len(data_tables) == 2
    assert data_tables[0].name == "X"
    assert data_tables[1].name == "layer1"

    converted_dataset = data_tables[0].data
    assert converted_dataset["id"] == [0, 1]
    assert converted_dataset["genes"] == [[1, 0, 2], [1, 2]]
    assert converted_dataset["expressions"] == [[1.0, 2.0, 3.0], [4.0, 6.0]]

    converted_dataset = data_tables[1].data
    assert converted_dataset["id"] == [0, 1]
    assert converted_dataset["genes"] == [[1, 0, 2], [1, 2]]
    assert converted_dataset["expressions"] == [[1.0, 2.0, 3.0], [4.0, 6.0]]



import pytest
import numpy as np
from unittest import mock
from anndata import AnnData
from pathlib import Path
import scipy.sparse as sp
from helical.models.geneformer.geneformer_tokenizer import TranscriptomeTokenizer, sum_ensembl_ids
from helical.constants.paths import CACHE_DIR_HELICAL

class TestGeneformerTokenizer:
    model_dir_v1 = Path(CACHE_DIR_HELICAL, "geneformer", "v1")
    model_dir_v2 = Path(CACHE_DIR_HELICAL, "geneformer", "v2")
    
    files_config_v1 = {
        "gene_median_path": model_dir_v1 / "gene_median_dictionary.pkl",
        "token_path": model_dir_v1 / "token_dictionary.pkl",
        "gene_mapping_path": model_dir_v1 / "ensembl_mapping_dict.pkl",
    }
    
    files_config_v2 = {
        "gene_median_path": model_dir_v2 / "gene_median_dictionary.pkl",
        "token_path": model_dir_v2 / "token_dictionary.pkl",
        "gene_mapping_path": model_dir_v2 / "ensembl_mapping_dict.pkl",
    }

    @pytest.fixture
    def tokenizer_v1(self):
        return TranscriptomeTokenizer(
            gene_median_file=self.files_config_v1["gene_median_path"],
            token_dictionary_file=self.files_config_v1["token_path"],
            gene_mapping_file=self.files_config_v1["gene_mapping_path"],
            model_input_size=2048,
            special_token=False,
        )

    @pytest.fixture
    def tokenizer_v2(self):
        return TranscriptomeTokenizer(
            gene_median_file=self.files_config_v2["gene_median_path"],
            token_dictionary_file=self.files_config_v2["token_path"],
            gene_mapping_file=self.files_config_v2["gene_mapping_path"],
            model_input_size=4096,
            special_token=True,
        )

    def test_tokenizer_initialization(self, tokenizer_v1, tokenizer_v2):
        assert tokenizer_v1.model_input_size == 2048
        assert not tokenizer_v1.special_token
        assert tokenizer_v2.model_input_size == 4096
        assert tokenizer_v2.special_token

    @pytest.mark.parametrize("collapse_gene_ids, expected_shape, expected_exception, data_format", [
    (True, (2, 2), None, 'h5ad'),
    (False, (2, 3), ValueError, 'h5ad'),
    (True, (2, 3), ValueError, 'bad_format')
    ])
    def test_sum_ensembl_ids_h5ad(self, collapse_gene_ids, expected_shape, expected_exception, data_format):
        # Create a test AnnData object
        adata = AnnData(X=sp.csr_matrix([[1, 2, 3], [4, 5, 6]]))
        adata.var['ensembl_id'] = ['ENSG1', 'ENSG2', 'ENSG2']
        adata.obs['n_counts'] = [6, 15]

        # Mock gene_mapping_dict and gene_token_dict
        gene_mapping_dict = {'ENSG1': 'ENSG1', 'ENSG2': 'ENSG2'}
        gene_token_dict = {'ENSG1': 1, 'ENSG2': 2}

        if expected_exception:
            with pytest.raises(expected_exception):
                sum_ensembl_ids(adata, collapse_gene_ids, gene_mapping_dict, gene_token_dict, file_format=data_format)
        else:
            result = sum_ensembl_ids(adata, collapse_gene_ids, gene_mapping_dict, gene_token_dict, file_format=data_format)
            assert isinstance(result, AnnData)
            assert result.shape == expected_shape

    def test_tokenize_anndata(self, tokenizer_v1):
        # Create a test AnnData object
        adata = AnnData(X=sp.csr_matrix([[1, 2, 3], [4, 5, 6]]))
        adata.var['ensembl_id'] = ['ENSG1', 'ENSG2', 'ENSG3']
        adata.obs['n_counts'] = [6, 15]
        adata.obs['cell_type'] = ['A', 'B']

        tokenizer_v1.custom_attr_name_dict = {'cell_type': 'cell_type'}
        tokenized_cells, cell_metadata = tokenizer_v1.tokenize_anndata(adata)

        assert len(tokenized_cells) == 2
        assert isinstance(tokenized_cells[0], np.ndarray)
        assert cell_metadata == {'cell_type': ['A', 'B']}

    @pytest.mark.parametrize("tokenizer, expected_max_length", [
        ("tokenizer_v1", 2048),
        ("tokenizer_v2", 4096)
    ])
    def test_create_dataset(self, request, tokenizer, expected_max_length):
        tokenizer = request.getfixturevalue(tokenizer)
        tokenized_cells = [np.array([1, 2, 3]), np.array([4, 5, 6])]
        cell_metadata = None

        dataset = tokenizer.create_dataset(tokenized_cells, cell_metadata)
        assert len(dataset) == 2
        assert 'input_ids' in dataset.features
        assert 'length' in dataset.features
        assert all(len(ids) <= expected_max_length for ids in dataset['input_ids'])

        if tokenizer.special_token:
            assert all(ids[0] == tokenizer.gene_token_dict['<cls>'] for ids in dataset['input_ids'])
            assert all(ids[-1] == tokenizer.gene_token_dict['<eos>'] for ids in dataset['input_ids'])

    def test_tokenize_data(self, tokenizer_v1, tmp_path):
        # Create a test AnnData object
        adata = AnnData(X=sp.csr_matrix([[1, 2, 3], [4, 5, 6]]))
        adata.var['ensembl_id'] = ['ENSG1', 'ENSG2', 'ENSG3']
        adata.obs['n_counts'] = [6, 15]
        adata.obs['cell_type'] = ['A', 'B']
        test_file = tmp_path / "test.h5ad"
        adata.write_h5ad(test_file)

        output_dir = tmp_path / "output"
        output_dir.mkdir()

        tokenizer_v1.tokenize_data(tmp_path, output_dir, "test_output", file_format="h5ad")

        assert (output_dir / "test_output.dataset").exists()

        with pytest.raises(ValueError):
            tokenizer_v1.tokenize_data(tmp_path, output_dir, "test_output", file_format="bad_format")




from helical import HelixmRNA, HelixmRNAConfig
import hydra
from omegaconf import DictConfig

@hydra.main(version_base=None, config_path="../run_models/configs", config_name="helix_mrna_config")
def run(cfg: DictConfig):
    input_sequences = ["EACU"*20, "EAUG"*20, "EAUG"*20, "EACU"*20, "EAUU"*20]

    helix_mrna_config = HelixmRNAConfig(**cfg)
    helix_mrna = HelixmRNA(helix_mrna_config)

    processed_input_data = helix_mrna.process_data(input_sequences)

    embeddings = helix_mrna.get_embeddings(processed_input_data)
    print(embeddings.shape)

if __name__ == "__main__":
    run()



import pytest
import torch
from helical import HyenaDNAConfig, HyenaDNAFineTuningModel

@pytest.mark.skip(reason="Work in progress.")
class TestHyenaDNAFineTuning:
    @pytest.fixture(params=["hyenadna-tiny-1k-seqlen", "hyenadna-tiny-1k-seqlen-d256"])
    def hyenaDNAFineTune(self, request):
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        config = HyenaDNAConfig(model_name=request.param, batch_size=1, device=self.device)
        return HyenaDNAFineTuningModel(hyena_config=config, fine_tuning_head="classification", output_size=1)

    @pytest.fixture
    def mock_data(self, hyenaDNAFineTune):
        input_sequences = ["AAAA", "CCCC", "TTTT", "ACGT", "ACGN", "BHIK", "ANNT"]
        labels = [0, 0, 0, 0, 0, 0, 0]
        tokenized_sequences = hyenaDNAFineTune.process_data(input_sequences)
        return tokenized_sequences, labels

    def test_output_dimensionality_of_fine_tuned_model(self, hyenaDNAFineTune, mock_data):
        input_sequences, labels = mock_data
        hyenaDNAFineTune.train(train_input_data=input_sequences, train_labels=labels, validation_input_data=input_sequences, validation_labels=labels)
        outputs = hyenaDNAFineTune.get_outputs(input_sequences)
        assert outputs.shape == (len(input_sequences), 1)


from helical.models.classification.neural_network import NeuralNetwork
from helical.models.classification.classifier import Classifier 
from helical.models.scgpt.model import scGPT
import pytest 
import anndata as ad
import numpy as np

def test_load_model__ok_nn_head():
    """
    Happy path for Classifier.load_model.
    The load_model expects a base model and a trained task model.
    - The base model needs to follow the BaseModelProtocol.
        In this case, the base model is an instance of scGPT. As all HelicalBaseFoundationModels have a process_data and 'get_embeddings' method, this is fine.
    - The trained task model needs to follow the ClassificationModelProtocol.
        In this case, the trained task model is an instance of NeuralNetwork. As the NeuralNetwork class inherits from a BaseTaskModel which requires the 
        implementation of the 'predict' method.
    """
    scgpt = scGPT()
    saved_nn_head = NeuralNetwork().load('ci/tests/data/valid_nn_head/my_model.h5', 'ci/tests/data/valid_nn_head/classes.npy')
    name = "scGPT with saved NN"
    scgpt_loaded_nn = Classifier().load_model(scgpt, saved_nn_head, name)    
    assert isinstance(scgpt_loaded_nn, Classifier)
    assert scgpt_loaded_nn.base_model == scgpt
    assert scgpt_loaded_nn.trained_task_model == saved_nn_head
    assert scgpt_loaded_nn.name == name
    assert scgpt_loaded_nn.gene_names == "index"


def test_load_model__ok_custom_head():
    """
    Happy path for Classifier.load_model.
    If a user decides to use his own model head, he just needs to make sure it is following the ClassificationModelProtocol.
    """
    scgpt = scGPT()

    class CustomModel:
        def __init__(self):
            pass
        def load(self):
            pass
        def predict(self):
            pass

    custom_head = CustomModel()
    name = "scGPT with custom NN"
    scgpt_loaded_nn = Classifier().load_model(scgpt, custom_head, name)    
    assert isinstance(scgpt_loaded_nn, Classifier)
    assert scgpt_loaded_nn.base_model == scgpt
    assert scgpt_loaded_nn.trained_task_model == custom_head
    assert scgpt_loaded_nn.name == name
    assert scgpt_loaded_nn.gene_names == "index"

def test_load_model__nok_custom_head():
    """
    Unappy path for Classifier.load_model.
    If a user decides to use his own model head, he just needs to make sure it is following the ClassificationModelProtocol.
    If this is not the case, raise an error.
    """
    scgpt = scGPT()

    class CustomModelHead:
        def __init__(self):
            pass
        # missing predict method

    custom_head = CustomModelHead()
    name = "scGPT with custom NN"
    with pytest.raises(TypeError):
        Classifier().load_model(scgpt, custom_head, name)    

def test_load_model__ok_custom_base():
    """
    Happy path for Classifier.load_model.
    If a user decides to use his own model head, he just needs to make sure it is following the ClassificationModelProtocol.
    """

    class CustomBaseModel:
        def __init__(self):
            pass
        def get_embeddings(self):
            pass
        def process_data(self):
            pass
        def train(self):
            pass

    custom_base = CustomBaseModel()

    saved_nn_head = NeuralNetwork().load('ci/tests/data/valid_nn_head/my_model.h5', 'ci/tests/data/valid_nn_head/classes.npy')
    name = "custom base model with NN"
    custom_base_classifier = Classifier().load_model(custom_base, saved_nn_head, name) 
    assert isinstance(custom_base_classifier, Classifier)
    assert custom_base_classifier.base_model == custom_base
    assert custom_base_classifier.trained_task_model == saved_nn_head
    assert custom_base_classifier.name == name
    assert custom_base_classifier.gene_names == "index"

def test_load_model__nok_custom_base():
    """
    Happy path for Classifier.load_model.
    If a user decides to use his own model head, he just needs to make sure it is following the ClassificationModelProtocol.
    """

    class CustomBaseModel:
        def __init__(self):
            pass
        # missing get_embeddings and process_data method

    custom_base = CustomBaseModel()

    saved_nn_head = NeuralNetwork().load('ci/tests/data/valid_nn_head/my_model.h5', 'ci/tests/data/valid_nn_head/classes.npy')
    name = "custom base model with NN"
    with pytest.raises(TypeError):
        Classifier().load_model(custom_base, saved_nn_head, name)    
    
def test_train_classifier_head(mocker):
    """
    Happy path for Classifier.train_classifier_head.
    """
    scgpt = scGPT()
    data = ad.read_h5ad("ci/tests/data/cell_type_sample.h5ad")
    head = NeuralNetwork()

    # Mocking the get_embeddings method by returning a zero matrix
    mocker.patch.object(scgpt, 'get_embeddings')
    scgpt.get_embeddings.return_value = np.zeros((data.shape[0], 10))
    
    # Mocking the training process by returning the untrained head
    mocker.patch.object(head, 'train')
    head.train.return_value = head

    scgpt_nn_c = Classifier().train_classifier_head(data, scgpt, head, labels_column_name = "cell_type")
    assert isinstance(scgpt_nn_c, Classifier)
    assert scgpt_nn_c.base_model == scgpt
    assert scgpt_nn_c.name == f"{scgpt.__class__.__name__} with {head.__class__.__name__}"
    assert scgpt_nn_c.trained_task_model == head 
    assert scgpt_nn_c.gene_names == "index"      


def test_check_validity_for_training_wrong_column_label_name():
    """
    Checks if a TypeError is raised if the labels_column_name is not present in the obs attribute of the AnnData object.
    """
    scgpt = scGPT()
    data = ad.read_h5ad("ci/tests/data/cell_type_sample.h5ad")
    with pytest.raises(TypeError):
        Classifier()._check_validity_for_training(data, labels_column_name = "wrong_name", base_model = scgpt) 


def test_check_validity_for_training_wrong_base_model():
    """
    Checks if a TypeError is raised if the base model is not an instance of a class implementing 'BaseModelProtocol'.
    """
    class CustomBaseModel:
        def __init__(self):
            pass
        # missing get_embeddings and process_data method

    custom_base = CustomBaseModel()
    data = ad.read_h5ad("ci/tests/data/cell_type_sample.h5ad")
    with pytest.raises(TypeError):
        Classifier()._check_validity_for_training(data, labels_column_name = "cell_type", base_model = custom_base) 


def test_get_predictions_w_no_base_model(mocker):
    """
    Happy path for Classifier.get_predictions having no base model.
    This is the case where the trained_task_model is a standalone classifier (handle anndata as input and predict classes as output).
    """

    base_model = None
    data = ad.read_h5ad("ci/tests/data/cell_type_sample.h5ad")
    standalone = NeuralNetwork().load('ci/tests/data/valid_nn_head/my_model.h5', 'ci/tests/data/valid_nn_head/classes.npy')

    # Mocking the prediction process by returning the untrained head
    mocker.patch.object(standalone, 'predict')
    standalone.predict.return_value = np.array(['type1', 'type2'] * (data.shape[0] // 2))

    name = "None as base with custom class as standalone classifier"
    c = Classifier().load_model(base_model, standalone, name)    
    c.get_predictions(data)
    assert isinstance(c, Classifier)
    assert c.base_model == None
    assert c.name == name
    assert c.trained_task_model == standalone   
    assert c.gene_names == "index"    